#define _LARGEFILE64_SOURCE

#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>

#include <sys/stat.h>
#include <unistd.h>

#include <argh.hh>

#include "binpacker.hh"

using namespace std;

#define STAT stat64
#define OFFT __off64_t

static int Exit(const string &message)
{
    fprintf(stderr, "Exit: %s\n", message.c_str());
    exit(-1);
}

class filespec
{
    string fn;
    OFFT size;
    bool ok;
public:
    filespec(const string &s) : fn(s)
    {
    }
    bool isok() const { return ok; }
    void GetSize()
    {
        struct STAT st;
        ok = !STAT(fn.c_str(), &st);
        if(ok)
        {
            if(!S_ISREG(st.st_mode))
            {
                fprintf(stderr, "'%s' is troublesome.\n", fn.c_str());
                Exit("file is not a file");
            }
            size = st.st_size;
        }
        else
            perror(fn.c_str());
    }
    OFFT getsize() const { return size; }
    const string &getname() const { return fn; }
};

typedef vector<filespec> filelist;

class cdspec
{
    const filelist &files;
    vector<unsigned> list;
public:
    cdspec(const filelist &tab) : files(tab)
    {
    }
    
    void operator<< (unsigned n)
    {
        files.at(n); // verify bounds
        list.insert(list.begin(), n);
    }
    void operator>> (unsigned n)
    {
        files.at(n); // verify bounds
        list.push_back(n);
    }
    void operator-= (unsigned n)
    {
        for(unsigned a=0; a<list.size(); )
            if(list[a] == n)
                erase(a);
            else
                ++a;
    }
    unsigned size() const { return list.size(); }
    unsigned operator[] (unsigned n) const { return list.at(n); }
    void erase(unsigned n) { list.erase(list.begin() + n); }
    void clear() { list.clear(); }
    
    OFFT getsize() const
    {
        OFFT size = 0;
        for(unsigned a=0; a<list.size(); ++a)
            size += files[list[a]].getsize();
        return size;
    }
};

typedef vector<cdspec> cdlist;

static OFFT ParseCdSize(const string &s)
{
    const char *q = s.c_str();
    double value = strtod(q, const_cast<char**>(&q));
    while(*q==' ')++q;
    switch(*q)
    {
        case 't':
        case 'T': value *= 1024.0; //fallthru
        case 'g':
        case 'G': value *= 1024.0; //fallthru
        case 'm':
        case 'M': value *= 1024.0; //fallthru
        case 'k': value *= 1024.0;
    }
    return static_cast<OFFT> (value);
}

static bool Check(const filelist &files, const cdlist &cd, OFFT cdsize)
{
    for(unsigned c=0; c<cd.size(); ++c)
    {
        printf("CD %u: %.1f MB", c, cd[c].getsize() / 1048576.0);
        printf(" (files");
        for(unsigned a=0; a<cd[c].size(); ++a)
            printf(" %u", cd[c][a]);
        printf(")");
        printf("\n");
    }
    
    bool needorder = false;
    for(unsigned c=0; c<cd.size(); ++c)
        if(cd[c].getsize() > cdsize)
            needorder = true;
    
    return !needorder;
}

static bool Algorithm1(const filelist &files, cdlist &cd, OFFT cdsize)
{
    unsigned cdcount = cd.size();
    for(unsigned a=0; a<cdcount; ++a) cd[a].clear();
    
    for(unsigned a=0, c=cdcount; a < files.size(); )
        for(--c; a < files.size(); ++a)
        {
            unsigned f = files.size() - a - 1;
            if(c)
            {
                OFFT size = cd[c].getsize();
                if(size + files[f].getsize() > cdsize)break;
            }
            cd[c] << f;
        }
    return Check(files, cd, cdsize);
}

static bool Algorithm2(const filelist &files, cdlist &cd, OFFT cdsize)
{
    unsigned cdcount = cd.size();
    for(unsigned a=0; a<cdcount; ++a) cd[a].clear();
    
    vector<OFFT> bins(cdcount);
    vector<OFFT> items(files.size());
    
    for(unsigned a=0; a<bins.size(); ++a) bins[a] = cdsize;
    for(unsigned a=0; a<items.size(); ++a) items[a] = files[a].getsize();
    
    vector<unsigned> result = PackBins(bins, items);
    
    for(unsigned a=0; a<items.size(); ++a) cd[result[a]] << a;
    
    return Check(files, cd, cdsize);
}

int main(int argc, const char *const *argv)
{
    ParamHandler Argh;
    
    OFFT cdsize = 650 << 20;
    
    filelist files;
    
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("cdsize",  'c').SetString().SetDesc("Specify CD size (650 MB default)");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");

    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'c':
                cdsize = ParseCdSize(Argh.GetString());
                break;
            case 'h':
                printf(
                    "jaa v"VERSION" - Copyright (C) 1992,2003 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\nUsage: jaa <option> [<...>]] <file> [<...>]\n"
                    "\nThis programs organizes files to CDs, floppies or such.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf("\nNo warranty whatsoever.\n");
                return 0;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    
    if(!files.size())
    {
        return Exit("No files specified");
    }
    
    for(unsigned a=0; a<files.size(); ++a)
        files[a].GetSize();

    for(unsigned a=0; a<files.size(); ++a)
        if(!files[a].isok())
        {
            return Exit("input file sizes unknown");
        }
    
    OFFT total = 0;
    for(unsigned a=0; a<files.size(); ++a)
    {
        if(files[a].getsize() > cdsize)
        {
            fprintf(stderr, "'%s' doesn't fit on any disc.\n", files[a].getname().c_str());
            return Exit("file bigger than any disc");
        }
        total += files[a].getsize();
    }
    
    unsigned cdcount = (total+cdsize-1) / cdsize;
    OFFT loss = cdcount * cdsize - total;
    
    printf("Number of CDs required at minimum: %u ", cdcount);
    printf(" (waste space total: %.1f MB)\n", loss / 1048576.0);
    
    cdlist cd(cdcount, cdspec(files));
    
    if(Algorithm1(files, cd, cdsize))
    {
        printf("Files need no scrambling. Just put them in order.\n");
        return 0;
    }
    printf("Direct method failed, trying another\n");
    if(Algorithm2(files, cd, cdsize))
    {
        printf("There.\n");
        return 0;
    }

    printf("Sorry, it didn't work out.\n");
    return 0;
}
